/* bi_objGL.c - A. Christidis*/
/*Binokulare Darstellung eines animierten Objektes mit OpenGL: 
Nach Vorgabe eines Sichtwinkels (field of view) fuer das gegebene 
Fenster wird die Konvergenz (conv) in die Projektionsebene gelegt
und das Achsenkreuz und ein Objekt dorthin (oder bei einer bel. 
Lage axBias[] geladen. Das Objekt wird dann entlang der z-Achse 
animiert (verschoben) und kann dabei um die 3 Achsen gedreht 
werden. Auch die Achsen selbst koennen (um ihren Ursprung bei 
axBias[]) gedreht werden. */

#define DOUBLEBUFF //Double Buffering
#define WITH_LIGHT
#define WORK_4TEA //Workaround fuer Teapot 

#include "bi_objGL.h"

enum {X=0, Y=1, Z=2, W=3};
enum {FALSE, TRUE};
enum {POINT, CUBE, TORUS, TEAPOT};

/*Globale Variablen: */
int      winWidth=512, winHeight=512; 
int      binoc=0, axes=1, aa=0, obj=0, backF=0, objRot=1;
float    fov=0.f, conv=0.f, disp=0.f, axBias[3];
int      shade=GL_FLAT, wire=GL_LINE;
GLfloat  pntVrtx[1][3], cubVrtx[8][3]; 
int      cubFace[6][4] = { {0, 1, 2, 3}, {1, 0, 4, 5}, {1, 5, 6, 2}, 
                           {4, 0, 3, 7}, {3, 2, 6, 7}, {5, 4, 7, 6} };
GLdouble left[]={0.,0.}, right[]={0.,0.}, bottom=0., top=0., near=0., far=0.;
GLfloat  pos[3]={0., 0., 0.}, angle[3]={0., 0., 0.}, axAngle[3]={0., 0., 0.}; 
char     anim=' ';

#ifdef  WITH_LIGHT
  int     light=0;
  GLfloat light_ambient[] = { 0.1, 0.1, 0.1, 1.0 };
  GLfloat light_diffuse[] = { 1.0, 1.0, 1.0, 1.0 };
  GLfloat light_specular[] = { 1.0, 1.0, 1.0, 1.0 };
  GLfloat light_position[] = { -1.0, 1.0, 1.0, 0.0 }; 
#endif//WITH_LIGHT

/*************************************************************************/
  void drawCube(void)
/*************************************************************************/
{ int     jj=0;

  /*Wuerfel zeichnen:*/
  for (jj = 0; jj < 6; jj++)
  {
    glBegin(GL_POLYGON); 
#ifdef  WITH_LIGHT
      glNormal3fv(cubVrtx[cubFace[jj][0]]);
#endif//WITH_LIGHT
    glVertex3fv(cubVrtx[cubFace[jj][0]]);
#ifdef  WITH_LIGHT
      glNormal3fv(cubVrtx[cubFace[jj][1]]);
#endif//WITH_LIGHT
    glVertex3fv(cubVrtx[cubFace[jj][1]]);
#ifdef  WITH_LIGHT
      glNormal3fv(cubVrtx[cubFace[jj][2]]);
#endif//WITH_LIGHT
    glVertex3fv(cubVrtx[cubFace[jj][2]]);
#ifdef  WITH_LIGHT
      glNormal3fv(cubVrtx[cubFace[jj][3]]);
#endif//WITH_LIGHT
    glVertex3fv(cubVrtx[cubFace[jj][3]]);
    glEnd();
  }
}

/*************************************************************************/
  void moveObj(void)
/*************************************************************************/
{ glMatrixMode(GL_MODELVIEW);

  /*Entlag der Achsen verschieben:*/
  glTranslatef(pos[X], pos[Y], pos[Z]);

  /*Drehen um Raumachsen:*/
  glRotatef(angle[Z], 0.0, 0.0, 1.0);
  glRotatef(angle[Y], 0.0, 1.0, 0.0);
  glRotatef(angle[X], 1.0, 0.0, 0.0);
}

/*************************************************************************/
  void drawObj(void)
/*************************************************************************/
{ int jj;

  /*Farbwahl, Linienstaerke:*/
  glColor3ub(255, 0,  255);
  glLineWidth (1.5);

  /*OpenGL-PolygonMode, -ShadeModel:*/
  glPolygonMode(GL_FRONT_AND_BACK, wire);
  glShadeModel (shade);

  /*Objekt zeichnen:*/
  switch (obj)
  { case POINT:  glPointSize (20.);  
                 for (jj=0; jj<1; jj++)
                 { glBegin(GL_POINTS);
                    glVertex3fv(pntVrtx[jj]);
                   glEnd(); } 
                 break;
    case CUBE:   drawCube();
                 break;
    case TORUS:  if (wire == GL_LINE) 
                      glutWireTorus(conv/20., 3*conv/20., 8, 10);
                 else glutSolidTorus(conv/20., 3*conv/20., 8, 10);
                 break;
    case TEAPOT:  
#ifdef  WORK_4TEA 
                 /*Workaround fuer Teapot - verliert sonst vorausgegangene Trfn.
                 z.B.: Starten mit Torus, Drehung um -195 um x, Wechsel auf 
                 Teapot: Mind. ein Farbauszug erscheint falsch.*/
                 moveAxes();
                 moveObj();
#endif//WORK_4TEA
                 glCallList(TEAPOT);
                 break;
  }
}
/*

/*************************************************************************/
  void moveAxes(void)
/*************************************************************************/
/*Zeigt geneigtes Achsenkreuz an der Stelle axBias*/
{
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity(); //Tabula rasa...

  glTranslatef(axBias[X], axBias[Y], axBias[Z]);

  /*Nicken, Gieren, Rollen (v.u.n.o.) vor der Verlegung:*/
  glRotatef(axAngle[Z], 0.0, 0.0, 1.0);
  glRotatef(axAngle[Y], 0.0, 1.0, 0.0);
  glRotatef(axAngle[X], 1.0, 0.0, 0.0);
}

/*************************************************************************/
  void drawAxes(void)
/*************************************************************************/
{ int   j1=0; float dummy=10.;
  float axCol[3][3]={{1.f,.2f,.0f}, {.0f,1.f,.0f}, {.0f,.6f,1.f}}; 
#ifdef  WITH_LIGHT
  /*Licht aus, da Strich ohne Materialeigenschaften weiss:*/
  if (light) { glDisable(GL_LIGHTING);  glDisable(GL_LIGHT0); }
 #endif//WITH_LIGHT

  /*Achsenkreuz:*/
  glLineWidth (0.8);  //glColor3f(.4,.4,.4);
  glBegin(GL_LINES);
   glColor3fv (axCol[X]); 
   glVertex3f (-far, 0.0, 0.0);   glVertex3f (far, 0.0, 0.0);
   glColor3fv (axCol[Y]); 
   glVertex3f (0.0, -far, 0.0);   glVertex3f (0.0, far, 0.0);
   glColor3fv (axCol[Z]); 
   glVertex3f (0., 0., near);   glVertex3f (0., 0., -far);
  glEnd();

#ifdef  WITH_LIGHT
  /*Licht ggf. wieder einschalten:*/
  if (light) { glEnable(GL_LIGHTING);  glEnable(GL_LIGHT0); }
#endif//WITH_LIGHT
}

/*************************************************************************/
  void draw(void)
/*************************************************************************/
{ int jj;

  /*Fenster loeschen (alle Farbkomponenten!):*/
  glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
  glClear(GL_COLOR_BUFFER_BIT);

#ifdef  WITH_LIGHT
  /*Tiefenpuffer fuer Achsen reinigen:*/
  glClear(GL_DEPTH_BUFFER_BIT);
#endif//WITH_LIGHT

  /*Achsen positionieren, ggf. zeichnen:*/
  moveAxes();
  if (axes) drawAxes();

  /*Vorm Weiterrechnen aktuelle Modell-Positionierung merken:*/
  glPushMatrix(); 
  /*Objekt positionieren:*/
  moveObj();

  /*Zeichnen (2x):*/
  for (jj=0; jj<2; jj++)
  { glMatrixMode(GL_PROJECTION);
    glLoadIdentity(); //Tabula rasa...

    /*Zweimal dieselbe Szene - zuvor Tiefenpuffer reinigen:*/
    glClear(GL_DEPTH_BUFFER_BIT);

    /*Augenpunkt positionieren:*/
    if (jj == 0) //links - rot
    { glColorMask(GL_TRUE, GL_FALSE, GL_FALSE, GL_TRUE);
      glFrustum (left[jj], right[jj], bottom, top, near, far);
      /*Positive x- fuer Links-Verschiebung des Augenpunkts:*/
      glTranslatef(disp/2., 0.0f, 0.0f);
    } else //rechts - blau
    { glColorMask(GL_FALSE, GL_FALSE, GL_TRUE, GL_TRUE); 
      glFrustum (left[jj], right[jj], bottom, top, near, far);
      /*Negative x- fuer Rechts-Verschiebung des Augenpunkts:*/
      glTranslatef(-disp/2., 0.0f, 0.0f);
    }

    /*Neu zeichnen:*/
    drawObj();
  }

  /*Modell vergessen, Achsen-Position behalten:*/
  glPopMatrix (); 

  /*Zeichenspeicher sichtbar machen bzw. Zeichnen erzwingen:*/
#ifdef  DOUBLEBUFF
  glutSwapBuffers();
#else //DOUBLEBUFF
  glFinish();
#endif//DOUBLEBUFF
}

/*************************************************************************/
  void animObj(int go)
/*************************************************************************/
{ static int   steps=100; //wechselndes Vorzeichen wiederfinden
         float fern=MIN(far,3.*conv), nahe=.4*conv;

  if (anim == 'G') 
  { draw(); 
    if (steps > 0) //entfernen
    { if (-pos[Z]-axBias[Z] < fern) pos[Z] -= (fern-nahe)/steps; 
      else steps = -steps; 
    } else //naeher holen
    { if (-pos[Z]-axBias[Z] > nahe) pos[Z] -= (fern-nahe)/steps; 
      else steps = -steps;
    }
    key(' ', 0, 0); //Menue aktualisieren
    glutTimerFunc(60,animObj,1);
  }
}

/*************************************************************************/
  void key(unsigned char key, int x, int y)
/*************************************************************************/
/*Menue und Eingabe-Behandlung:*/
{ switch (key)
  { case ESC: exit(0);
    case CR : objRot = 1-objRot;      break;
    case '+': axes = 1-axes;          break;
    case '0':
    case '1':
    case '2':
    case '3': obj = key - '0';
              if (obj == TEAPOT) 
              { teapot();//Bildung der Koordinatenliste
                /* Teapot-Flaechenpunkte im Uhrzeigersinn:*/
                glFrontFace(GL_CW); }
              else glFrontFace(GL_CCW);
              break;

    case 'a': aa = 1 - aa;  if (aa)  /*Anti-Aliasing:*/
              { glEnable (GL_POINT_SMOOTH); 
                glEnable (GL_LINE_SMOOTH);  glEnable (GL_BLEND); 
                glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
              } else 
              { glDisable (GL_POINT_SMOOTH); 
                glDisable (GL_LINE_SMOOTH);  glDisable (GL_BLEND);
              }                       break;
    case 'b': backF = 1-backF; 
              if (backF) glEnable  (GL_CULL_FACE);
              else       glDisable (GL_CULL_FACE);
              break;
    case 'g': anim = key;  break; 
    case 'G': if (anim == 'G') break; 
              else { anim = key; animObj(1);  break; }

#ifdef  WITH_LIGHT
    case 'l': light = 1-light;
              if (light) { glEnable(GL_LIGHTING);
                           glEnable(GL_LIGHT0);
                           shade = GL_SMOOTH; 
                           glEnable(GL_DEPTH_TEST);
                         } else
                         { glDisable(GL_LIGHTING);
                           glDisable(GL_LIGHT0);
                           shade = GL_FLAT;  
                           glDisable(GL_DEPTH_TEST);
                         }
              break;
#endif//WITH_LIGHT

    case 'm': binoc = 1-binoc;
              if (binoc) disp=conv/30.; else disp = 0.; 
              right[0] = near * (winWidth + disp) / (2*conv); 
              right[1] = near * (winWidth - disp) / (2*conv); 
              left[0] = - right[1];  left[1] = - right[0];  
                                      break;
    case 'r': angle[X]=angle[Y]=angle[Z]=0.;
              axAngle[X]=axAngle[Y]=axAngle[Z]=0.;
              pos[Z] = 0;             break;
    case 'S': pos[Z] -= .1*(pos[Z]+axBias[Z]);         break;
    case 'W': pos[Z] += .1*(pos[Z]+axBias[Z]);         break;
    case 'w': wire  = GL_LINE+GL_FILL - wire; 
              if (wire == GL_FILL) 
              { backF = 1;  glEnable(GL_CULL_FACE); }
              if (obj == TEAPOT) teapot(); //Bildung  Koordinatenliste
              break;

    case 'x': if (objRot) 
              { angle[X] -= 5.; 
                if (angle[X] <=-360) angle[X]+=360; break; } else
              { axAngle[X] -= 5.; 
                if (axAngle[X] <=-360) axAngle[X]+=360; break; }
    case 'X': if (objRot)
              { angle[X] += 5.; 
                if (angle[X] >= 360) angle[X]-=360; break; } else 
                { axAngle[X] += 5.; 
                  if (axAngle[X] >= 360) axAngle[X]-=360; break; }
    case 'y': if (objRot)
              { angle[Y] -= 5.; 
                if (angle[Y] <=-360) angle[Y]+=360; break; } else
              { axAngle[Y] -= 5.; 
                if (axAngle[Y] <=-360) axAngle[Y]+=360; break; }
    case 'Y': if (objRot)
              { angle[Y] += 5.; 
                if (angle[Y] >= 360) angle[Y]-=360; break; } else
              { axAngle[Y] += 5.; 
                if (axAngle[Y] >= 360) axAngle[Y]-=360; break; }
    case 'z': if (objRot)
              { angle[Z] -= 5.; 
                if (angle[Z] <=-360) angle[Z]+=360; break; } else
              { axAngle[Z] -= 5.; 
                if (axAngle[Z] <=-360) axAngle[Z]+=360; break; }
    case 'Z': if (objRot)
              { angle[Z] += 5.; 
                if (angle[Z] >= 360) angle[Z]-=360; break; } else
              { axAngle[Z] += 5.; 
                if (axAngle[Z] >= 360) axAngle[Z]-=360; break; }
  }
  system (CON_CLS);
  printf ("\n Press (keeping the GLUT window activated):\n");
  printf ("\n <ESC> to quit");
  printf ("\n <CR>     toggle rotate object "); if (objRot) printf ("(ON)");
  printf ("/ coord.sys. "); if (!objRot) printf ("(ON)");
  printf ("\n 0/1/2/3  choose point"); if (obj == POINT) printf (" (ON)");
  printf ("/cube"); if (obj == CUBE) printf (" (ON)");
  printf ("/torus"); if (obj == TORUS) printf (" (ON)");
  printf ("/teapot"); if (obj == TEAPOT) printf (" (ON)");
  printf ("\n +        toggle show coordinate axes ");
  if (axes) printf ("(ON)"); else printf ("(OFF)"); 
  printf ("\n a        toggle Line & Point AntiAliasing ");
  if (aa) printf ("(ON)"); else printf ("(OFF)"); 
  printf ("\n b        toggle Backface culling ");
  if (backF) printf ("(ON)"); else printf ("(OFF)"); 
  printf ("\n G / g    give automatic animation a Go ");
  if (anim == 'G') printf ("(ON)"); else printf ("(OFF)"); 
#ifdef  WITH_LIGHT
  printf ("\n l        toggle Light ");
  if (light) printf ("(ON)"); else printf ("(OFF)"); 
#endif//WITH_LIGHT
  printf ("\n m        toggle Monocular "); 
  if (!binoc) printf ("(ON)");
  printf ("/ binocular "); if (binoc) printf ("(ON)");
  printf ("\n r        Reload");
  printf ("\n S / W    decrease/increase Z by 10%% (curr.: %.2f)",pos[Z]+axBias[Z]);
  printf ("\n w        toggle PolygonMode: Wire "); 
  if (wire == GL_LINE) printf ("(ON)");
  printf ("/ fill "); if (wire == GL_FILL) printf ("(ON)");
  printf ("\n x / X    decrease/increase X-rotation angle ");
  if (objRot) printf ("object"); else printf ("coord.sys. "); 
  printf ("\n y / Y    decrease/increase Y-rotation angle ");
  if (objRot) printf ("object"); else printf ("coord.sys. "); 
  printf ("\n z / Z    decrease/increase Z-rotation angle ");
  if (objRot) printf ("object \n"); else printf ("coord.sys. \n"); 
  printf ("\n Current values: convergence=%8.2f units", -conv);
  printf ("\n               field of view=%8.2f deg.", fov);
  printf ("\n        object angles[X,Y,Z]=[%7.2f,%7.2f,%7.2f] deg.", angle[X], angle[Y], angle[Z]);
  printf ("\n  coord.system angles[X,Y,Z]=[%7.2f,%7.2f,%7.2f] deg.", axAngle[X], axAngle[Y], axAngle[Z]);
  draw();
  return;
}

/*************************************************************************/
  void init(void)
/*************************************************************************/
  /*Initialisierung globaler Variablen und Konstanten bei Deklaration 
  nur mit Zahlenwert moeglich, in Funktionen auch durch Berechnung:*/
{ const  double  GRAD  = atan(1.) / 45.;
  float cubSize;
  fov = 60.f;
  conv = winWidth/(2*tan(fov*GRAD/2.));
  far = 100.*conv;
  near = conv/8.; 
  right[0] = near * (winWidth + disp) / (2*conv); 
  right[1] = near * (winWidth - disp) / (2*conv); 
  left[0] = - right[1];  left[1] = - right[0];  
  top  = near * winHeight/ (2*conv);  bottom = -top;
  axBias[Z]=-conv;
  pos[Z] = 0; 

  /*Punkt-Koordinaten:*/
  pntVrtx[0][X] = pntVrtx[0][Y] = pntVrtx[0][Z] = 0.;

  /*Wuerfel-Koordinten:*/
  cubSize=conv/8.;
  cubVrtx[0][X] = cubVrtx[3][X] = cubVrtx[4][X] = cubVrtx[7][X] = -cubSize;
  cubVrtx[1][X] = cubVrtx[2][X] = cubVrtx[5][X] = cubVrtx[6][X] =  cubSize;
  cubVrtx[0][Y] = cubVrtx[1][Y] = cubVrtx[4][Y] = cubVrtx[5][Y] = -cubSize;
  cubVrtx[2][Y] = cubVrtx[3][Y] = cubVrtx[6][Y] = cubVrtx[7][Y] =  cubSize;
  cubVrtx[0][Z] = cubVrtx[1][Z] = cubVrtx[2][Z] = cubVrtx[3][Z] =  cubSize;
  cubVrtx[4][Z] = cubVrtx[5][Z] = cubVrtx[6][Z] = cubVrtx[7][Z] = -cubSize;


#ifdef  WITH_LIGHT
  glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
  glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
  glLightfv(GL_LIGHT0, GL_POSITION, light_position);
#endif//WITH_LIGHT

  /*Tastendruck simulieren, um Menue auszugeben:*/
  key(' ', 0, 0);
}
/*************************************************************************/
  void teapot(void)
/*************************************************************************/
/*Koordinatenliste Utah-Teapot als Display List:*/
{ glNewList(TEAPOT, GL_COMPILE); //Liste anlegen

  if (wire == GL_LINE) glutWireTeapot(conv/8.0);
  else                 glutSolidTeapot(conv/8.0);

  glEndList();   //Fertig
}

/*************************************************************************/
  int main(int argc, char **argv)
/*************************************************************************/
{ glutInit(&argc, argv);

#ifdef  WITH_LIGHT
glutInitDisplayMode (GLUT_DEPTH);
#endif//WITH_LIGHT

#ifdef  DOUBLEBUFF
  glutInitDisplayMode(GLUT_DOUBLE);
#endif//DOUBLEBUFF
  glutInitWindowSize(winWidth, winHeight);
  glutCreateWindow("SwTP goes OpenGL");
  glutDisplayFunc(draw);
  glutKeyboardFunc(key);
  init(); 
  glutMainLoop();
  return 0;
}